#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# This file is part of the `plugy` python module
#
# Copyright
# 2018-2021
# EMBL & Heidelberg University
#
# Author(s): Dénes Türei (turei.denes@gmail.com)
#            Nicolas Peschke
#            Olga Ivanova
#
# Distributed under the GPLv3 License.
# See accompanying file LICENSE.txt or copy at
#     http://www.gnu.org/licenses/gpl-3.0.html
#
# Webpage: https://github.com/saezlab/plugy
#


import unittest
import tempfile

import logging
import gzip
import pathlib as pl

import numpy as np

import pandas as pd
import pandas.testing as pd_test

from ..data import pmt
from ..data import config

FILE_CONTENT_COMMA = """LabVIEW Measurement
Writer_Version\t2
Reader_Version\t2
Separator\tTab
Decimal_Separator\t,
Multi_Headings\tNo
X_Columns\tNo
Time_Pref\tRelative
Operator\tEveline
Date\t2019/07/22
Time\t17:18:18,938945770263671875
***End_of_Header***

Channels\t4
Samples\t100\t100\t100\t100
Date\t2019/07/22\t2019/07/22\t2019/07/22\t2019/07/22
Time\t17:18:18,938945770263671875\t17:18:18,938945770263671875\t17:18:18,938945770263671875\t17:18:18,938945770263671875
X_Dimension\tTime\tTime\tTime\tTime
X0\t0,0000000000000000E+0\t0,0000000000000000E+0\t0,0000000000000000E+0\t0,0000000000000000E+0
Delta_X\t1,000000\t1,000000\t1,000000\t1,000000
***End_of_Header***
X_Value\tUntitled\tUntitled 1\tUntitled 2\tUntitled 3\tComment
\t0,000000\t0,055544\t0,032960\t0,071718
\t0,000000\t0,054323\t0,032044\t0,049745
\t0,000000\t0,055239\t0,032655\t0,050050
\t0,000000\t0,053713\t0,031739\t0,049135
\t0,000000\t0,055544\t0,032655\t0,048830
\t0,000000\t0,055849\t0,033265\t0,050356
"""

FILE_CONTENT_DOT = """LabVIEW Measurement
Writer_Version\t2
Reader_Version\t2
Separator\tTab
Decimal_Separator\t.
Multi_Headings\tNo
X_Columns\tNo
Time_Pref\tRelative
Operator\tEveline
Date\t2019/07/22
Time\t17:18:18
***End_of_Header***

Channels\t4
Samples\t100\t100\t100\t100
Date\t2019/07/22\t2019/07/22\t2019/07/22\t2019/07/22
Time\t17:18:18\t17:18:18\t17:18:18\t17:18:18
X_Dimension\tTime\tTime\tTime\tTime
X0\t0.0000000000000000E+0\t0.0000000000000000E+0\t0.0000000000000000E+0\t0.0000000000000000E+0
Delta_X\t1.000000\t1.000000\t1.000000\t1.000000
***End_of_Header***
X_Value\tUntitled\tUntitled 1\tUntitled 2\tUntitled 3\tComment
\t0.000000\t0.055544\t0.032960\t0.071718
\t0.000000\t0.054323\t0.032044\t0.049745
\t0.000000\t0.055239\t0.032655\t0.050050
\t0.000000\t0.053713\t0.031739\t0.049135
\t0.000000\t0.055544\t0.032655\t0.048830
\t0.000000\t0.055849\t0.033265\t0.050356
"""

logging.basicConfig(level = logging.DEBUG,
                    format = '%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                    datefmt = '%d.%m.%y %H:%M:%S')


class TestPmtData(unittest.TestCase):


    def setUp(self):
        """
        Creates DataFrame for each unittest to compare results
        """
        self.test_df = pd.DataFrame({
            "time": [
                0.000000, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000
            ],
            "green": [
                0.055544, 0.054323, 0.055239, 0.053713, 0.055544, 0.055849
            ],
            "orange": [
                0.032960, 0.032044, 0.032655, 0.031739, 0.032655, 0.033265
            ],
            "uv": [
                0.071718, 0.049745, 0.050050, 0.049135, 0.048830, 0.050356
            ]
        })

        self.tmpdir = tempfile.mkdtemp()
        self.pmt_path = pl.Path(self.tmpdir, 'exp.txt')
        self.pmt_path.touch()
        self.config = config.PlugyConfig(input_dir = self.tmpdir)


    def test_gz_file_open(self):
        """
        Checks if reading gz file returns the expected DataFrame
        """
        file_contents = [FILE_CONTENT_COMMA, FILE_CONTENT_DOT]

        for file_content in file_contents:

            with self.subTest(file_content = file_content):

                with tempfile.NamedTemporaryFile(
                        mode = "w+b",
                        suffix = ".txt.gz",
                        delete = True
                    ) as self.gz_file:

                    with gzip.GzipFile(mode = "wb", fileobj = self.gz_file) as gz:

                        gz.write(file_content.encode())

                    self.gz_file.seek(0)
                    self.gz_file_path = pl.Path(self.gz_file.name)

                    self.data = pmt.PmtData(
                        self.gz_file_path,
                        config = self.config,
                    ).read_txt()

                pd_test.assert_frame_equal(self.test_df, self.data)


    def test_txt_file_open(self):
        """
        Checks if reading normal txt file returns the expected DataFrame
        """

        file_contents = [FILE_CONTENT_COMMA, FILE_CONTENT_DOT]

        for file_content in file_contents:

            with self.subTest(file_content = file_content):

                with tempfile.NamedTemporaryFile(
                        mode = "w+t",
                        suffix = ".txt",
                        delete = True
                    ) as self.txt_file:

                    self.txt_file.write(file_content)
                    self.txt_file.seek(0)
                    self.txt_file_path = pl.Path(self.txt_file.name)

                    self.data = pmt.PmtData(
                        self.txt_file_path,
                        config = self.config,
                    ).read_txt()

                pd_test.assert_frame_equal(self.test_df, self.data)


    def test_other_file_open(self):
        """
        Checks error handling in read_txt
        """

        self.suffix = ".any"

        with tempfile.NamedTemporaryFile(suffix = self.suffix) as self.any_file:

            self.any_file_path = pl.Path(self.any_file.name)

            with self.assertRaises(NotImplementedError) as cm:

                data = pmt.PmtData(
                    self.any_file_path,
                    config = self.config,
                ).read_txt()

            self.assertEqual(
                cm.exception.args[0],
                f"Input file has to be either .txt or .txt.gz, "
                f"{self.any_file_path.suffix} files are not implemented!"
            )


    def test_find_data(self):
        """
        Checks error handling in find_data
        """

        # Checking exception raise with empty file
        with tempfile.TemporaryFile() as self.empty_file:

            with self.assertRaises(AssertionError) as cm:

                pmt.PmtData.find_data(self.empty_file)

            self.assertEqual(
                cm.exception.args[0],
                "No lines detected in input_file! "
                "Check the contents of the file!"
            )

        # Checking error raised with too long header
        with tempfile.TemporaryFile(mode = "w+t") as self.wrong_header_file:

            self.wrong_header_file.writelines(["test\n" for _ in range(100)])
            self.wrong_header_file.seek(0)

            with self.assertRaises(AssertionError) as cm:

                pmt.PmtData.find_data(self.wrong_header_file)

        # Checking with real header (comma separated)
        with tempfile.TemporaryFile(mode = "w+t") as self.right_file_comma:

            self.right_file_comma.write(FILE_CONTENT_COMMA)
            self.right_file_comma.seek(0)

            self.assertEqual(pmt.PmtData.find_data(self.right_file_comma), 22)

        # Checking with real header (dot separated)
        with tempfile.TemporaryFile(mode = "w+t") as self.right_file_dot:
            self.right_file_dot.write(FILE_CONTENT_DOT)
            self.right_file_dot.seek(0)

            self.assertEqual(pmt.PmtData.find_data(self.right_file_dot), 22)


    def test_detect_decimal_separator(self):
        """
        Checks if decimal separator is detected properly
        """

        with tempfile.TemporaryFile(mode = "w+t") as self.right_file_dot:
            self.right_file_dot.write(FILE_CONTENT_DOT)
            self.right_file_dot.seek(0)

            self.assertEqual(
                pmt.PmtData.detect_decimal_separator(self.right_file_dot),
                "."
            )

        with tempfile.TemporaryFile(mode = "w+t") as self.right_file_comma:

            self.right_file_comma.write(FILE_CONTENT_COMMA)
            self.right_file_comma.seek(0)

            self.assertEqual(
                pmt.PmtData.detect_decimal_separator(self.right_file_comma),
                ","
            )

        with tempfile.TemporaryFile(mode = "w+t") as self.wrong_header_file:
            self.wrong_header_file.writelines(["test\n" for _ in range(100)])
            self.wrong_header_file.seek(0)

            self.assertEqual(
                pmt.PmtData.detect_decimal_separator(self.wrong_header_file),
                ","
            )


    # noinspection PyArgumentList
    def test_set_channel_value_ignore(self):
        """
        Tests ignoring individual channels
        """

        test_df_zero_green = self.test_df.assign(green = 0.0)
        test_df_zero_uv = self.test_df.assign(uv = 0.0)
        test_df_zero_orange = self.test_df.assign(orange = 0.0)

        with unittest.mock.patch.object(
            target = pmt.PmtData,
            attribute = "read_txt",
            new = lambda _: self.test_df
        ):

            with self.subTest():

                data = pmt.PmtData(
                    pl.Path(),
                    correct_acquisition_time = False,
                    ignore_channels = {'green'},
                    config = self.config,
                ).data

                pd_test.assert_frame_equal(data, test_df_zero_green)

                data = pmt.PmtData(
                    pl.Path(),
                    correct_acquisition_time = False,
                    ignore_channels = {'orange'},
                    config = self.config,
                ).data

                pd_test.assert_frame_equal(data, test_df_zero_orange)

                data = pmt.PmtData(
                    pl.Path(),
                    correct_acquisition_time = False,
                    ignore_channels = {'uv'},
                    config = self.config,
                ).data

                pd_test.assert_frame_equal(data, test_df_zero_uv)


    # noinspection PyArgumentList
    def test_set_channel_value_time(self):
        """
        Tests correcting the time values
        """

        with unittest.mock.patch.object(
            target = pmt.PmtData,
            attribute = "read_txt",
            new = lambda _: self.test_df,
        ):

            for acq in range(100, 500, 100):

                with self.subTest(acq = acq):

                    test_df_time = self.test_df.assign(
                        time = np.linspace(
                            0,
                            (1 / acq) * (len(self.test_df) - 1),
                            len(self.test_df)
                        )
                    )

                    data = pmt.PmtData(
                        pl.Path(),
                        correct_acquisition_time = True,
                        acquisition_rate = acq,
                        config = self.config,
                    ).data

                    pd_test.assert_frame_equal(data, test_df_time)

            test_df_time = self.test_df.assign(
                time = np.linspace(
                    0,
                    len(self.test_df) - 1,
                    len(self.test_df)
                )
            )
            data = pmt.PmtData(
                pl.Path(),
                correct_acquisition_time = True,
                acquisition_rate = 1,
                config = self.config,
            ).data

            pd_test.assert_frame_equal(data, test_df_time)

            data = pmt.PmtData(
                pl.Path(),
                correct_acquisition_time = False,
                config = self.config,
            ).data

            pd_test.assert_frame_equal(data, self.test_df)


    def test_cut_data(self):
        """
        Tests if input is cut properly
        """

        with (
            unittest.mock.patch.object(
                target = pmt.PmtData,
                attribute = "read_txt",
                new = lambda _: self.test_df
            )
        ):

            cut = (1, 4)
            data = pmt.PmtData(
                input_file = pl.Path(),
                acquisition_rate = 1,
                correct_acquisition_time = True,
                cut = cut,
                config = self.config,
            ).data
            self.assertTrue(len(data.time) == 4)
            self.assertTrue(min(data.time) >= cut[0])
            self.assertTrue(max(data.time) <= cut[1])

            cut = (None, None)
            data = pmt.PmtData(
                input_file = pl.Path(),
                acquisition_rate = 1,
                correct_acquisition_time = True,
                cut = cut,
                config = self.config,
            ).data
            self.assertTrue(len(data.time) == len(self.test_df))

            cut = (3, None)
            data = pmt.PmtData(
                input_file = pl.Path(),
                acquisition_rate = 1,
                correct_acquisition_time = True,
                cut = cut,
                config = self.config,
            ).data
            self.assertTrue(len(data.time) == 3)
            self.assertTrue(min(data.time) >= cut[0])

            cut = (None, 3)
            data = pmt.PmtData(
                input_file = pl.Path(),
                acquisition_rate = 1,
                correct_acquisition_time = True,
                cut = cut,
                config = self.config,
            ).data
            self.assertTrue(len(data.time) == 4)
            self.assertTrue(max(data.time) <= cut[1])

            cut = (4, 1)
            with self.assertRaises(AttributeError) as cm:

                data = pmt.PmtData(
                    input_file = pl.Path(),
                    acquisition_rate = 1,
                    correct_acquisition_time = True,
                    cut = cut,
                    config = self.config,
                ).data

            self.assertEqual(
                cm.exception.args[0],
                f"Cut has to be specified like cut = (min, max) "
                f"you specified {cut}"
            )


    # noinspection DuplicatedCode
    def test_cut_additional_data(self):
        """
        Tests if cutting of supplied df works opposed to using
        the df in the PmtData object.
        """

        with (
            unittest.mock.patch.object(
                target = pmt.PmtData,
                attribute = "read_txt",
                new = lambda _: self.test_df
            )
        ):

            raw_data = pmt.PmtData(
                input_file = pl.Path(),
                acquisition_rate = 1,
                correct_acquisition_time = True,
                cut = (None, None),
                config = self.config,
            )

        cut = (1, 4)
        data = raw_data.cut_data(cut = cut)
        self.assertTrue(len(data.time) == 4)
        self.assertTrue(min(data.time) >= cut[0])
        self.assertTrue(max(data.time) <= cut[1])

        cut = (None, None)
        data = raw_data.cut_data(cut = cut)
        self.assertTrue(len(data.time) == len(self.test_df))

        cut = (3, None)
        data = raw_data.cut_data(cut = cut)
        self.assertTrue(len(data.time) == 3)
        self.assertTrue(min(data.time) >= cut[0])

        cut = (None, 3)
        data = raw_data.cut_data(cut = cut)
        self.assertTrue(len(data.time) == 4)
        self.assertTrue(max(data.time) <= cut[1])

        cut = (4, 1)
        with self.assertRaises(AttributeError) as cm:

            data = raw_data.cut_data(cut = cut)

        self.assertEqual(
            cm.exception.args[0],
            f"Cut has to be specified like cut = (min, max) "
            f"you specified {cut}"
        )


    def test_fake_gain(self):
        """
        Tests digital gain method
        """

        with unittest.mock.patch.object(
            target = pmt.PmtData,
            attribute = "read_txt",
            new = lambda _: self.test_df,
        ):

            data = pmt.PmtData(
                input_file = pl.Path(),
                acquisition_rate = 1,
                correct_acquisition_time = False,
                fake_gains = {'uv': 2},
                config = self.config,
            ).data

            pd_test.assert_frame_equal(
                data,
                self.test_df.assign(
                    uv = [
                        0.143436, 0.09949, 0.1001,
                        0.09827, 0.09766, 0.100712
                    ]
                )
            )

            data = pmt.PmtData(
                input_file = pl.Path(),
                acquisition_rate = 1,
                correct_acquisition_time = False,
                fake_gains = {'green': 2},
                config = self.config,
            ).data

            pd_test.assert_frame_equal(
                data,
                self.test_df.assign(
                    green = [
                        0.111088, 0.108646, 0.110478,
                        0.107426, 0.111088, 0.111698
                    ]
                )
            )

            data = pmt.PmtData(
                input_file = pl.Path(),
                acquisition_rate = 1,
                correct_acquisition_time = False,
                fake_gains = {'orange': 2},
                config = self.config,
            ).data

            pd_test.assert_frame_equal(
                data,
                self.test_df.assign(
                    orange = [
                        0.06592, 0.064088, 0.06531,
                        0.063478, 0.06531, 0.06653
                    ]
                )
            )


if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# This file is part of the `plugy` python module
#
# Copyright
# 2018-2021
# EMBL & Heidelberg University
#
# Author(s): Dénes Türei (turei.denes@gmail.com)
#            Nicolas Peschke
#            Olga Ivanova
#
# Distributed under the GPLv3 License.
# See accompanying file LICENSE.txt or copy at
#     http://www.gnu.org/licenses/gpl-3.0.html
#
# Webpage: https://github.com/saezlab/plugy
#

import logging
import tempfile
import unittest

import pathlib as pl

from ..data.bd import ChannelMap, PlugSequence, Sample

logging.basicConfig(level = logging.DEBUG,
                    format = '%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                    datefmt = '%d.%m.%y %H:%M:%S')


class TestChannelMapping(unittest.TestCase):
    def setUp(self) -> None:
        """
        Prepare dictionary with channel mapping and produce string of the dict to write to file
        """
        self.test_mapping = {9: "CELLS", 10: "SUBSTRATE", 11: "FS", 12: "FS"}
        for idx, i in enumerate(range(13, 23)):
            self.test_mapping[i] = f"Drug {idx + 1}"
        self.test_mapping[23] = "BCL"
        self.test_mapping[24] = "BCH"

        self.test_file_content = "\n".join([f"{k}:{v}" for k, v in self.test_mapping.items()])

    def test_read_input_file_mapping(self):
        """
        Test if a file with the correct contents is read properly
        """
        with tempfile.NamedTemporaryFile(mode = "w+t", suffix = ".txt") as self.channel_file:
            self.channel_file.write(self.test_file_content)
            self.channel_file.seek(0)
            self.channel_file_path = pl.Path(self.channel_file.name)

            mapping = ChannelMap(input_file = self.channel_file_path)

        self.assertEqual(self.test_mapping, mapping.map)

    def test_read_input_file_bc(self):
        """
        Test if a file with the correct contents is read properly
        """
        with tempfile.NamedTemporaryFile(mode = "w+t", suffix = ".txt") as self.channel_file:
            self.channel_file.write(self.test_file_content)
            self.channel_file.seek(0)
            self.channel_file_path = pl.Path(self.channel_file.name)

            mapping = ChannelMap(input_file = self.channel_file_path)

        self.assertEqual([23, 24], mapping.bc)

    def test_read_input_file_drugs(self):
        """
        Test if a file with the correct contents is read properly
        """
        with tempfile.NamedTemporaryFile(mode = "w+t", suffix = ".txt") as self.channel_file:
            self.channel_file.write(self.test_file_content)
            self.channel_file.seek(0)
            self.channel_file_path = pl.Path(self.channel_file.name)

            mapping = ChannelMap(input_file = self.channel_file_path)

        self.assertEqual([i for i in range(13, 23)], mapping.drugs)

    def test_read_input_file_cells(self):
        """
        Test if a file with the correct contents is read properly
        """
        with tempfile.NamedTemporaryFile(mode = "w+t", suffix = ".txt") as self.channel_file:
            self.channel_file.write(self.test_file_content)
            self.channel_file.seek(0)
            self.channel_file_path = pl.Path(self.channel_file.name)

            mapping = ChannelMap(input_file = self.channel_file_path)

        self.assertEqual([9], mapping.cells)

    def test_read_input_file_substrate(self):
        """
        Test if a file with the correct contents is read properly
        """
        with tempfile.NamedTemporaryFile(mode = "w+t", suffix = ".txt") as self.channel_file:
            self.channel_file.write(self.test_file_content)
            self.channel_file.seek(0)
            self.channel_file_path = pl.Path(self.channel_file.name)

            mapping = ChannelMap(input_file = self.channel_file_path)

        self.assertEqual([10], mapping.substrate)

    def test_read_input_file_media(self):
        """
        Test if a file with the correct contents is read properly
        """
        with tempfile.NamedTemporaryFile(mode = "w+t", suffix = ".txt") as self.channel_file:
            self.channel_file.write(self.test_file_content)
            self.channel_file.seek(0)
            self.channel_file_path = pl.Path(self.channel_file.name)

            mapping = ChannelMap(input_file = self.channel_file_path)

        self.assertEqual([11, 12], mapping.media)

    def test_read_broken_input_file_channel(self):
        """
        Tests if a channel not on the braille chip is raising a ValueError
        """
        with tempfile.NamedTemporaryFile(mode = "w+t", suffix = ".txt") as self.channel_file:
            self.channel_file.write("1:Test")
            self.channel_file.seek(0)
            self.channel_file_path = pl.Path(self.channel_file.name)

            with self.assertRaises(ValueError) as cm:
                mapping = ChannelMap(self.channel_file_path)

        self.assertEqual(cm.exception.args[0], f"Channel out of BD range (9-24) you specified channel 1")

    def test_read_broken_input_file_channel_nan(self):
        """
        Tests if there is an ValueError raised when channel is something else than an int
        """
        with tempfile.NamedTemporaryFile(mode = "w+t", suffix = ".txt") as self.channel_nan_file:
            self.channel_nan_file.write("ABC:Test")
            self.channel_nan_file.seek(0)
            self.channel_nan_file_path = pl.Path(self.channel_nan_file.name)

            with self.assertRaises(ValueError) as cm:
                mapping = ChannelMap(self.channel_nan_file_path)

        self.assertEqual(cm.exception.args[0], "Channel has to be an int you specified ABC")


class TestPlugSequenceReadWrite(unittest.TestCase):
    def setUp(self) -> None:
        """
        Creates the test_file_content and a corresponding test_sequence that contains the true data
        """
        self.test_file_content = "\n1,12,Valve9,12,14,16,9\n1,12,Valve10,12,14,16,10\n1,12,Valve11,12,14,16,11\n"

        self.test_sequence = (Sample(1, 12, "Valve9", [12, 14, 16, 9]),
                              Sample(1, 12, "Valve10", [12, 14, 16, 10]),
                              Sample(1, 12, "Valve11", [12, 14, 16, 11]))

    def test_from_csv_file(self):
        """
        Test if a csv file with the correct contents is read properly
        """
        with tempfile.NamedTemporaryFile(mode = "w+t", suffix = ".csv") as self.sequence_file:
            self.sequence_file.write(self.test_file_content)
            self.sequence_file.seek(0)
            self.sequence_file_path = pl.Path(self.sequence_file.name)

            plug_sequence = PlugSequence.from_csv_file(input_file = self.sequence_file_path)

        self.assertEqual(self.test_sequence, plug_sequence.sequence)

    def test_save_csv_file(self):
        """
        Tests writing sequence to a csv file
        """
        plug_sequence = PlugSequence(self.test_sequence)
        with tempfile.NamedTemporaryFile(mode = "w+t", suffix = ".csv") as self.save_file:
            plug_sequence.save_csv(pl.Path(self.save_file.name))
            self.save_file.seek(0)
            self.assertEqual(self.test_file_content, self.save_file.read())


class TestPlugSequenceGenerate(unittest.TestCase):
    def setUp(self) -> None:
        self.test_gen_map_content = "9:CELLS\n10:SUBSTRATE\n11:FS\n12:FS\n13:Drug 1\n14:Drug 2\n15:Drug 3\n23:BCL\n24:BCH"
        self.test_sequence = (Sample(1, 15, "Cycle Barcode", [11, 12, 23, 24]),
                              Sample(1, 12, "Cell Control", [9, 10, 11, 12]),
                              Sample(1, 10, "Barcode", [11, 12, 23, 24]),
                              Sample(1, 12, "Drug 1", [9, 10, 11, 13]),
                              Sample(1, 10, "Barcode", [11, 12, 23, 24]),
                              Sample(1, 12, "Drug 2", [9, 10, 11, 14]),
                              Sample(1, 10, "Barcode", [11, 12, 23, 24]),
                              Sample(1, 12, "Drug 3", [9, 10, 11, 15]),
                              Sample(1, 10, "Barcode", [11, 12, 23, 24]),
                              Sample(1, 12, "Cell Control", [9, 10, 11, 12]),
                              Sample(1, 10, "Barcode", [11, 12, 23, 24]),
                              Sample(1, 12, "Drug 1 + Drug 2", [9, 10, 13, 14]),
                              Sample(1, 10, "Barcode", [11, 12, 23, 24]),
                              Sample(1, 12, "Drug 1 + Drug 3", [9, 10, 13, 15]),
                              Sample(1, 10, "Barcode", [11, 12, 23, 24]),
                              Sample(1, 12, "Drug 2 + Drug 3", [9, 10, 14, 15]),
                              Sample(1, 10, "Barcode", [11, 12, 23, 24]),
                              Sample(1, 12, "Cell Control", [9, 10, 11, 12]),
                              Sample(1, 10, "Barcode", [11, 12, 23, 24]),
                              Sample(1, 15, "Cycle Barcode", [11, 12, 23, 24]))

    def test_from_channel_map(self):
        with tempfile.NamedTemporaryFile(mode = "w+t", suffix = ".txt") as self.channel_file:
            self.channel_file.write(self.test_gen_map_content)
            self.channel_file.seek(0)
            self.channel_map = ChannelMap(pl.Path(self.channel_file.name))

        plug_sequence = PlugSequence.from_channel_map(self.channel_map, n_replicates = 12, n_control = 12, n_barcode = 10, n_cycle_bc = 15, open_duration = 1, generate_barcodes = True)
        self.assertEqual(self.test_sequence, plug_sequence.sequence)


class TestPlugSequenceCheck(unittest.TestCase):
    """
    Tests sanity checking of the PlugSequence input
    """
    def test_sequence_type(self):
        with self.assertRaises(TypeError) as cm:
            PlugSequence((Sample(1, 12, "Test", [11, 12, 13, 14]), "test"))
        self.assertEqual(cm.exception.args[0], "Samples in the plug sequence have to be of class Sample, you specified <class 'str'> in sample 1")

    def test_sequence_many_valves(self):
        with self.assertRaises(ValueError) as cm:
            PlugSequence((Sample(1, 12, "Test", [11, 12, 13, 14]), Sample(1, 12, "Test", [11, 12, 13, 14, 15])))
        self.assertEqual(
            cm.exception.args[0],
            "Sample 1 found with more than 4 valves open (5), "
            "THIS MIGHT DAMAGE THE CHIP!"
        )

    def test_sequence_few_valves(self):
        with self.assertWarns(UserWarning) as cm:
            PlugSequence((Sample(1, 12, "Test", [11, 12, 13, 14]), Sample(1, 12, "Test", [11, 12, 13])))
        self.assertEqual(cm.warning.args[0], "Less than 4 valves open (3) in sample 1")

    def test_sequence_other_valves(self):
        with self.assertRaises(ValueError) as cm:
            PlugSequence((Sample(1, 12, "Test", [11, 12, 13, 14]), Sample(1, 12, "Test", [1, -12, 13, 14])))
        self.assertEqual(cm.exception.args[0], "Sample 1 contains valves that are not used on the chip ([1, -12, 13, 14])")


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# This file is part of the `plugy` python module
#
# Copyright
# 2018-2021
# EMBL & Heidelberg University
#
# Author(s): Dénes Türei (turei.denes@gmail.com)
#            Nicolas Peschke
#            Olga Ivanova
#
# Distributed under the GPLv3 License.
# See accompanying file LICENSE.txt or copy at
#     http://www.gnu.org/licenses/gpl-3.0.html
#
# Webpage: https://github.com/saezlab/plugy
#


import collections as coll
import csv
import itertools
import collections
import logging
import pathlib as pl
import warnings
from typing import Union
from dataclasses import dataclass

from .. import misc

module_logger = logging.getLogger("plugy.data.bd")

Sample = coll.namedtuple("Sample", ["open_duration", "n_replicates", "name", "open_valves"])


@dataclass
class ChannelMap(object):
    input_file: Union[pl.Path, str]
    label_mapping: dict = None

    def __post_init__(self):

        self.input_file = pl.Path(self.input_file)
        module_logger.info(
            f"Creating ChannelMap object from "
            f"file {self.input_file.absolute()}"
        )
        module_logger.debug(f"Configuration:")
        for k, v in self.__dict__.items():
            module_logger.debug(f"{k}: {v}")

        self.label_startswith = {
            'cells': {'cell'},
            'substrate': {'sub'},
            'bc': {'bc', 'barc'},
            'media': {'fs', 'frees', 'medi'},
        }
        self.label_mapping = dict(
            (k, misc.to_set(v))
            for k, v in (self.label_mapping or {}).items()
        )

        self.cells = list()
        self.substrate = list()
        self.media = list()
        self.bc = list()
        self.drugs = list()

        self.map = self.read_input_file()

    # noinspection PyAttributeOutsideInit
    def read_input_file(self):
        module_logger.info(f"Reading file {self.input_file.absolute()}")
        mapping = dict()
        with self.input_file.open("r") as f:
            for line in f:
                k, v = line.strip(' \n\r\t;,').split(":")

                try:
                    k = int(k)
                except ValueError:
                    raise ValueError(
                        f"Channel has to be an int you specified {k}"
                    )

                if k not in range(9, 25):
                    raise ValueError(
                        f"Channel out of BD range (9-24) "
                        f"you specified channel {k}"
                    )

                channel_type = self._match_channel_label(v)

                if hasattr(self, channel_type):

                    getattr(self, channel_type).append(k)

                else:

                    raise ValueError(
                        f"Could not find out what type of channel is {str(v)}"
                        f". Please rename the channel or define the channel "
                        f"type in the `label_mapping` argument."
                    )

                mapping[k] = v

        return mapping


    def _match_channel_label(self, label):

        label = str(label)

        for channel_type, kws in self.label_mapping.items():

            if label in kws:

                return channel_type

        for channel_type, kws in self.label_startswith.items():

            for kw in kws:

                if label.lower().startswith(kw):

                    return channel_type

        return 'drugs'


    def get_compounds(self, open_valves):
        compounds = list()

        for open_valve in open_valves:
            if (open_valve in self.drugs) or (open_valve in self.media):
                compounds.append(self.map[open_valve])

            # if open_valves in self.cells:
            #     cells = True
            # if open_valves in self.substrate:
            #     substrate = True
            # if open_valves in self.media:
            #     media = True
            # if open_valves in self.bc:
            #     bc = True
            #
            #     drugs = True

        return compounds

    def get_compound_list(self):
        """
        Retrieves a list of the compounds used in the experiment including
        the media
        :return: list containing the names of the compounds
        """
        compounds = list()

        compounds.extend([self.map[valve] for valve in self.media])
        compounds.extend([self.map[valve] for valve in self.drugs])

        return compounds


class PlugSequence(object):

    @classmethod
    def from_csv_file(cls, input_file: Union[pl.Path, str], **kwargs):
        """
        Reads a "Samples on Demand v5" compatible csv file and creates a
            PlugSequence object from it.
        :param input_file: File path to read from
        :return: PlugSequence object
        """

        input_file = pl.Path(input_file)
        module_logger.info(f"Reading PlugSequence from {input_file.absolute()}")
        sequence = list()
        with input_file.open("r") as f:
            reader = csv.reader(f)
            for row in reader:
                if len(row) == 0:
                    continue
                else:
                    # noinspection PyCallByClass
                    sequence.append(
                        Sample(
                            open_duration=float(row[0]),
                            n_replicates=int(row[1]),
                            name=row[2],
                            open_valves=[int(i) for i in row[3:]]
                        )
                    )

        return cls(sequence=tuple(sequence), **kwargs)

    # noinspection PyCallByClass
    @classmethod
    def from_channel_map(
            cls,
            channel_map: ChannelMap,
            n_replicates: int = 12,
            n_control: int = 12,
            n_barcode: int = 5,
            n_cycle_bc: int = 15,
            open_duration: int = 1,
            generate_barcodes: bool = True,
            barcode_drug: int = None,
            **kwargs
        ):
        """
        Generates a PlugSequence object from a ChannelMap
        :param channel_map: ChannelMap to generate the PlugSequence from
        :param n_replicates: Number of replicates for drug combinations
        :param n_control: Number of replicates for the cell only controls
        :param n_barcode: Number of barcodes between samples
        :param n_cycle_bc: Number of barcodes at the beginning and end of a cylcle
        :param open_duration: How long the braille valves should be opened
            per plug in seconds
        :param generate_barcodes: If barcodes should be encoded in the
            Sequence. Set to False if barcodes should be generated using
            "Samples on Demand v5"
        :return: PlugSequence object
        """
        module_logger.info(f"Creating PlugSequence from ChannelMap")
        module_logger.debug("Configuration:")
        for k, v in locals().items():
            module_logger.debug(f"{k}: {v}")
        samples = list()

        # Generate templates
        control = Sample(
            open_duration = open_duration,
            n_replicates = n_control,
            name = "Cell Control",
            open_valves = (
                channel_map.cells +
                channel_map.substrate +
                channel_map.media
            )
        )
        barcode = Sample(
            open_duration = open_duration,
            n_replicates = n_barcode,
            name = "Barcode",
            open_valves = channel_map.media + channel_map.bc
        )
        cycle_bc = Sample(
            open_duration = open_duration,
            n_replicates = n_cycle_bc,
            name = "Cycle Barcode",
            open_valves = channel_map.media + channel_map.bc
        )

        if len(channel_map.bc) == 1:
            barcode_substitute = barcode_drug or channel_map.drugs[0]
            module_logger.warning(
                'Single barcode channel available, using `%s` instead.' %
                channel_map.map[barcode_substitute]
            )
            barcode.open_valves.append(barcode_substitute)
            cycle_bc.open_valves.append(barcode_substitute)

        individual_drugs = list()
        for drug in channel_map.drugs:
            individual_drugs.append(
                Sample(
                    open_duration = open_duration,
                    n_replicates = n_replicates,
                    name = channel_map.map[drug],
                    open_valves = (
                        channel_map.cells +
                        channel_map.substrate +
                        [channel_map.media[0]] + [drug]
                    )
                )
            )
            if generate_barcodes:
                individual_drugs.append(barcode)

        # Generate sample list
        samples.append(cycle_bc)
        samples.append(control)
        if generate_barcodes:
            samples.append(barcode)
        samples = samples + individual_drugs

        for idx, combination in enumerate(
            itertools.combinations(channel_map.drugs, 2)
        ):
            if idx % 10 == 0:
                samples.append(control)
                if generate_barcodes:
                    samples.append(barcode)

            samples.append(
                Sample(
                    open_duration = open_duration,
                    n_replicates = n_replicates,
                    name = (
                        f"{channel_map.map[combination[0]]} + "
                        f"{channel_map.map[combination[1]]}"
                    ),
                    open_valves = (
                        channel_map.cells +
                        channel_map.substrate +
                        list(combination)
                    )
                )
            )
            if generate_barcodes:
                samples.append(barcode)

        samples.append(control)
        if generate_barcodes:
            samples.append(barcode)

        samples.append(cycle_bc)
        return cls(sequence=tuple(samples), channel_map=channel_map, **kwargs)

    def __init__(
            self,
            sequence: tuple,
            allow_lt4_valves: bool = False,
            **kwargs
        ):
        """
        Handles the plug sequence of the braille display
        """
        self.sequence = sequence
        self.allow_lt4_valves = allow_lt4_valves
        self.check_sequence()

        try:
            if isinstance(kwargs["channel_map"], ChannelMap):
                self.channel_map = kwargs["channel_map"]
        except KeyError:
            pass

    def check_sequence(self):
        """
        Tests sanity of sequence
        """
        module_logger.debug("Checking plug sequence")

        for idx, sample in enumerate(self.sequence):
            if not isinstance(sample, Sample):
                raise TypeError(
                    f"Samples in the plug sequence have to be of class "
                    f"Sample, you specified {type(sample)} in sample {idx}"
                )

            if len(sample.open_valves) < 4 and not self.allow_lt4_valves:
                warnings.warn(
                    f"Less than 4 valves open ({len(sample.open_valves)}) "
                    f"in sample {idx}"
                )

            elif len(sample.open_valves) > 4:
                raise ValueError(
                    f"Sample {idx} found with more than 4 valves open "
                    f"({len(sample.open_valves)}), "
                    f"THIS MIGHT DAMAGE THE CHIP!"
                )

            for valve in sample.open_valves:
                if valve not in range(9, 25):
                    raise ValueError(
                        f"Sample {idx} contains valves that are "
                        f"not used on the chip ({sample.open_valves})"
                    )

    def save_csv(self, path: Union[str, pl.Path]):
        """
        Saves the PlugSequence as a csv file that is compatible to "Samples on Demand v5"
        :param path: Path to write to
        """
        path = pl.Path(path)
        module_logger.info(f"Writing sequence to file {path.absolute()}")
        with path.open("w", newline="\r\n") as f:
            f.write("\n")
            for sample in self.sequence:
                f.write(
                    f"{str(sample.open_duration)},"
                    f"{str(sample.n_replicates)},"
                    f"{str(sample.name)},"
                    f"{','.join([str(i) for i in sample.open_valves])}"
                )
                f.write("\n")

    def get_samples(self, **kwargs):
        """
        Filters barcodes out of the plug sequence and returns a PlugSequence
            object without barcodes
        :param kwargs: channel_map: optionally overrides the ChannelMap object
            that might be already present in the PlugSequence object
        :return: tuple containing Samples filtered
        """

        if "channel_map" in kwargs.keys():
            if isinstance(kwargs["channel_map"], ChannelMap):
                c_map = kwargs["channel_map"]
            else:
                raise TypeError(
                    f"channel_map of type {type(kwargs['channel_map'])} but "
                    f"has to be type ChannelMap"
                )
        else:
            try:
                c_map = self.channel_map
            except AttributeError:
                raise AttributeError(
                    "No ChannelMap object specified or found to get sample "
                    "information from! You can specify one with the "
                    "channel_map keyword argument."
                )

        filtered_samples = list()

        for sample in self.sequence:
            bc = False
            # if c_map.cells in sample.open_valves:
            #     cells.append(True)
            # else:
            #     cells.append(False)
            #
            # if c_map.substrate in sample.open_valves:
            #     substrate.append(True)
            # else:
            #     substrate.append(False)

            # for valve in sample.open_valves:
            #     if valve in c_map.drugs:
            #         filtered_samples.append(sample)
            #         continue

            for bc_valve in c_map.bc:
                if bc_valve in sample.open_valves:
                    bc = True
                    break

            if not bc:
                filtered_samples.append(sample)

        return PlugSequence(tuple(filtered_samples))


class RotaryValveSequence(object):


    def __init__(
            self,
            infile: str,
            channel_map: ChannelMap = None,
            oil_valve: Union[str,int] = None,
        ):

        self.infile = infile
        self.channel_map = channel_map
        self.oil_valve = oil_valve
        self.raw = []
        self.sequence = []
        self.main()


    def main(self):

        self.read()
        self.process()


    def read(self):

        RotaryValveStep = collections.namedtuple(
            'RotaryValveStep',
            ['start_time', 'length', 'flow_rate', 'valve']
        )
        t = 0
        seq = []

        module_logger.info(
            'Reading rotary valve sequence from `%s`' % self.infile
        )

        with open(self.infile, 'r') as fp:

            _ = fp.readline()

            for line in fp:

                line = line.strip().split(';')

                length = int(line[0])

                if line[2]:

                    seq.append(
                        RotaryValveStep(
                            start_time = t,
                            length = length,
                            flow_rate = int(line[1]),
                            valve = line[2],
                        )
                    )

                t += length

        self.raw = seq


    def process(self):

        RotaryValveSample = collections.namedtuple(
            'RotaryValveSample',
            [
                'start_time',
                'end_time',
                'lengths',
                'flow_rates',
                'valves',
                'components',
            ],
        )

        self.sequence = []

        for step in self.raw:

            pass


#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# This file is part of the `plugy` python module
#
# Copyright
# 2018-2021
# EMBL & Heidelberg University
#
# Author(s): Dénes Türei (turei.denes@gmail.com)
#            Nicolas Peschke
#            Olga Ivanova
#
# Distributed under the GPLv3 License.
# See accompanying file LICENSE.txt or copy at
#     http://www.gnu.org/licenses/gpl-3.0.html
#
# Webpage: https://github.com/saezlab/plugy
#


from setuptools import setup


with open('requirements.txt', 'r') as fp:

    requirements = [l.strip() for l in fp]

with open('README.md', 'r') as f:

    readme = f.read()


setup(
    name = 'plugy',
    version = '1.0.9',
    packages = ['plugy', 'plugy.data', 'plugy.test'],
    url = 'https://github.com/saezlab/plugy',
    license = 'GPLv3',
    author = 'Dénes Türei, Nicolas Peschke, Olga Ivanova',
    author_email = 'turei.denes@gmail.com',
    description = 'Processing plug microfluidics data',
    long_description = readme,
    long_description_content_type = 'text/markdown; charset=UTF-8',
    install_requires = requirements,
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# This file is part of the `plugy` python module
#
# Copyright
# 2018-2021
# EMBL & Heidelberg University
#
# Author(s): Dénes Türei (turei.denes@gmail.com)
#            Nicolas Peschke
#            Olga Ivanova
#
# Distributed under the GPLv3 License.
# See accompanying file LICENSE.txt or copy at
#     http://www.gnu.org/licenses/gpl-3.0.html
#
# Webpage: https://github.com/saezlab/plugy
#


import logging
import re
import gzip
import pathlib as pl
import importlib as imp
from typing import Union
from dataclasses import dataclass, field

import pandas as pd
import numpy as np
import scipy.signal as sig

import matplotlib.ticker as mpl_ticker
import matplotlib.pyplot as plt
import seaborn as sns

from ..data.config import PlugyConfig
from .. import misc

module_logger = logging.getLogger(__name__)


@dataclass
class PmtData(object):

    input_file: pl.Path
    acquisition_rate: int = 300
    cut: tuple = (None, None)
    correct_acquisition_time: bool = True

    channels: dict = field(
        default_factory = lambda: {
            'barcode': ('uv', 3),
            'control': ('orange', 2),
            'readout': ('green', 1),
        }
    )
    ignore_channels: set = field(default_factory = set)
    fake_gains: dict = field(default_factory = dict)
    fake_gain_default: float = 1.0
    fake_gain_adaptive: bool = False
    barcode_raw_threshold: float = None
    channels: dict = field(
        default_factory = lambda: {
            'barcode': ('uv', 3),
            'control': ('orange', 2),
            'readout': ('green', 1),
        }
    )
    peak_min_threshold: float = 0.05
    peak_max_threshold: float = 2.0
    peak_min_distance: float = 0.03
    peak_min_prominence: float = 0
    peak_max_prominence: float = 10
    peak_min_width: float = 0.5
    peak_max_width: float = 1.5
    width_rel_height: float = 0.5
    merge_peaks_distance: float = 0.2
    merge_peaks_by: str = 'center'

    config: PlugyConfig = field(default_factory = PlugyConfig)


    def __post_init__(self):

        module_logger.info(
            f"Creating PmtData object from file {self.input_file.absolute()}"
        )
        module_logger.debug(f"Configuration:")
        for k, v in self.__dict__.items():
            module_logger.debug(f"{k}: {v}")

        self.data = self.read_txt()
        self.data = self.set_channel_values()
        self.data = self.cut_data()

        self._set_fake_gain_adaptive()
        self.apply_fake_gain()
        self._override_barcode()


    def reload(self):
        """
        Reloads the object from the module level.
        """

        modname = self.__class__.__module__
        mod = __import__(modname, fromlist = [modname.split('.')[0]])
        imp.reload(mod)
        new = getattr(mod, self.__class__.__name__)
        setattr(self, '__class__', new)


    def read_txt(self) -> pd.DataFrame:
        """
        Reads input_file
        :return: pd.DataFrame containing the PMT data of all channels
        """
        module_logger.info(f"Reading file {self.input_file.absolute()}")
        if self.input_file.exists():
            if self.input_file.suffix == ".gz":
                module_logger.info("Detected gzipped file")
                with gzip.open(self.input_file, "rt") as f:
                    header_info = self.parse_header(f)

            elif self.input_file.suffix == ".txt":
                module_logger.info("Detected uncompressed txt file")
                with self.input_file.open("rt") as f:
                    header_info = self.parse_header(f)

            else:
                raise NotImplementedError(f"Input file has to be either .txt or .txt.gz, {self.input_file.suffix} files are not implemented!")

            data_frame = pd.read_csv(self.input_file, sep = "\t", decimal = header_info["decimal_separator"], skiprows = header_info["end_of_header"], header = None).iloc[:, 1:]
            data_frame.columns = ["time", "green", "orange", "uv"]

            return data_frame

        else:
            raise FileNotFoundError(f"Input file ({self.input_file.absolute()}) does not exist! Check the path!")


    def parse_header(self, file) -> dict:
        """
        Parses the LabView header and extracts information in form of a dict
        "end_of_header" : Line number of the first data line
        "decimal_separator" : Type of decimal separator
        :param file: read opened text file object
        :return: dictionary containing end_of_header and decimal_separator
        """
        info = dict()
        info["end_of_header"] = self.find_data(file)
        info["decimal_separator"] = self.detect_decimal_separator(file)
        return info


    @staticmethod
    def detect_decimal_separator(file) -> str:
        """
        Parses the LabView header and extracts the decimal separator
        :param file: read opened text file object
        :return: String containing the decimal separator
        """
        file.seek(0)
        for line in file:
            if line.startswith("Decimal_Separator"):
                sep = line.split()[1]
                module_logger.debug(f"Detected decimal separator: {sep}")
                return sep

        module_logger.error("Automatic decimal separator detection failed, falling back to ',' as decimal separator.")
        return ","

    @staticmethod
    def find_data(file) -> int:
        """
        Finds the ending of the header in a multichannel acquisition output file.
        Identifies data by its leading \t
        :param file: File object
        :return: Line number of the first data line
        """
        file.seek(0)
        idx = -1
        for idx, line in enumerate(file):
            if re.match(pattern = r"\t\d", string = line) is not None:
                break

        module_logger.debug(f"Detected end of header in line {idx}")
        assert idx > -1, "No lines detected in input_file! Check the contents of the file!"
        assert idx < 50, f"Automatically detected header length exceeds 50 lines ({idx})"

        return idx


    def cut_data(self, **kwargs) -> pd.DataFrame:
        """
        Returns data between time range specified in cut
        :param kwargs: "cut": specify an upper and lower limit (lower, upper) other than the one in the object already
        :return: pd.DataFrame containing the data in the time range
        """
        module_logger.debug(f"Cutting {self.__repr__()}")

        if "cut" in kwargs.keys():
            cut = kwargs["cut"]
        else:
            cut = self.cut

        df = self.data

        try:
            if cut[0] >= cut[1]:
                raise AttributeError(f"Cut has to be specified like cut = (min, max) you specified {cut}")
        except TypeError:
            # in case of comparison with None
            pass

        if cut[0] is not None:
            module_logger.debug(f"Cutting data before t = {cut[0]}")
            df = df.loc[df.time >= cut[0]]

        if cut[1] is not None:
            module_logger.debug(f"Cutting data after t = {cut[1]}")
            df = df.loc[df.time <= cut[1]]

        return df


    def set_channel_values(self) -> pd.DataFrame:
        """
        Sets & corrects values in the multichannel acquisition data.
        :return: DataFrame with the corrected data
        """
        time_between_samplings = 1 / self.acquisition_rate

        df = self.data.copy()

        for channel in (
            set(self.config.channel_names.values()) &
            misc.to_set(self.ignore_channels)
        ):

            module_logger.info('Setting %s channel to 0.0' % channel)
            df[channel] = .0

        if self.correct_acquisition_time:

            module_logger.info("Correcting acquisition time")
            df = df.assign(
                time = np.linspace(
                    self.data.time[0],
                    self.data.time[0] +
                        time_between_samplings *
                        (len(df) - 1),
                    len(df)
                )
            )

        return df


    def _set_fake_gain_adaptive(self):
        # for channel in ["uv", "green", "orange"]:
        #     pass

        if self.fake_gain_adaptive:

            raise NotImplementedError


    def apply_fake_gain(self):
        """
        Multiplies the corresponding channels PMT output by a factor provided
        in ``fake_gains``.
        """
        df = self.data

        for channel in self.config.channel_names.values():

            fake_gain = (
                self.fake_gains[channel]
                    if channel in self.fake_gains else
                self.fake_gain_default
            )

            if fake_gain != 1.0:

                module_logger.info(
                    f"Applying fake gain for {channel} channel: {fake_gain}"
                )
                df[channel] = df[channel] * fake_gain

        self.data = df


    def plot_pmt_data(
            self,
            axes: plt.Axes,
            cut: tuple = (None, None),
            ylim: tuple = (None, None),
            n_x_ticks: Union[int, float] = None,
        ) -> plt.Axes:
        """
        Plots the raw PMT data to the specified axes object.

        :param axes:
            plt.Axes object to draw on
        :param cut:
            Tuple to specify upper and lower time bounds for the pmt
            data to be plotted (lower, upper)

        :return:
            The axes object with the plot
        """
        module_logger.debug('Plotting PMT data')
        df = self.cut_data(cut = cut)

        with sns.axes_style(
            'darkgrid',
            {
                'xtick.bottom': True,
                'xtick.major.size': 1.0,
                'xtick.minor.size': 0.5
            },
        ):

            for channel in ('green', 'orange', 'uv'):

                sns.lineplot(
                    x = df.time,
                    y = df[channel],
                    estimator = None,
                    ci = None,
                    sort = False,
                    color = self.config.colors[channel],
                    linewidth = 1.,
                    ax = axes,
                )

            if df.time.max() - df.time.min() > 1000:

                major_tick_freq = 100
                minor_tick_freq = 25

            else:

                major_tick_freq = 10
                minor_tick_freq = 1

            axes.set_xticks(
                range(
                    int(round(df.time.min())),
                    int(round(df.time.max())),
                    major_tick_freq
                ),
                minor = False,
            )

            axes.set_xticks(
                range(
                    int(round(df.time.min())),
                    int(round(df.time.max())),
                    minor_tick_freq
                ),
                minor = True,
            )

            axes.set_xlim(
                left = int(round(df.time.min())),
                right = int(round(df.time.max())),
            )

            axes.set_ylim(ylim)

            if n_x_ticks:

                major_locator = plt.MaxNLocator(
                    nbins = n_x_ticks,
                    prune = 'lower',
                )
                minor_locator = plt.MaxNLocator(
                    nbins = n_x_ticks * 10,
                    prune = 'lower',
                )
                minor_locator.MAXTICKS = 10000

                axes.xaxis.set_major_locator(major_locator)
                axes.xaxis.set_minor_locator(minor_locator)

            axes.tick_params(
                axis = 'x',
                direction = 'out',
                length = 8, width = 2,
                bottom = True,
                which = 'major',
            )
            axes.tick_params(
                axis = 'x',
                direction = 'out',
                length = 4, width = 1,
                bottom = True,
                which = 'minor',
            )

            axes.grid(b = False, which = "major", axis = 'x')
            #axes.grid(b = True, which = "minor", color = "k", linewidth = 0.5)

            axes.set_facecolor("white")
            axes.tick_params(labelsize = "xx-large")

            axes.set_xlabel(
                "Time [s]",
                horizontalalignment = "left",
                size = "xx-large",
                x = 0,
            )
            axes.set_ylabel("Fluorescence [AU]", size = "xx-large")

        return axes


    def plot_peak_sequence(
            self,
            axes: plt.Axes,
            cut: tuple = (None, None),
        ) -> plt.Axes:
        """
        Creates a compact plot of the detected peaks: x axis is the sequence
        of the peaks, y axis is the median intensity of the channels.
        """

        module_logger.debug('Plotting peak sequence')
        df = self.peak_df

        for channel in ('barcode', 'control', 'readout'):

            if (
                self.config.channel_names[channel] in
                self.config.ignore_channels
            ):

                continue

            sns.scatterplot(
                x = np.arange(1, len(df) + 1),
                y = df['%s_peak_median' % channel],
                color = self.config.channel_color(channel),
                ax = axes,
                s = 50,
            )

        axes.set_facecolor('white')
        axes.grid(b = True, which = 'major', color = 'k', linewidth = 1.0)
        axes.grid(b = True, which = 'minor', color = 'k', linewidth = 0.5)

        if len(df) > 100:

            axes.xaxis.set_major_locator(mpl_ticker.MultipleLocator(20))

        axes.xaxis.set_minor_locator(
            mpl_ticker.MultipleLocator(5 if len(df) > 50 else 2)
        )
        axes.tick_params(labelsize = 'xx-large')
        axes.set_xlabel('Plug sequence', size = 'x-large')
        axes.set_ylabel('Fluorescence [AU]', size = 'x-large')
        axes.set_xlim((-5, len(df) + 5))

        return axes


    def correct_crosstalk(self) -> pd.DataFrame:
        """
        Implements crosstalk compensation as in Federica's BraDiPluS package
        (peaksSelection.R).

        :return: pd.DataFrame with corrected orange channel
        """
        # df = self.data
        #
        # df = df.assign(orange = df.orange - (0.45 * df.green))
        #
        # return df
        raise NotImplementedError


    def _override_barcode(self):
        """
        Overrides values in the barcode channel to 1 if current barcode
        value > bc_override_threshold.
        """

        if self.barcode_raw_threshold:

            module_logger.warning(
                f"Setting barcode channel to 1 for "
                f"barcode values > {self.barcode_raw_threshold}"
            )

            barcode = self.config.channels["barcode"][0]

            self.data[barcode].loc[
                self.data[barcode] > self.barcode_raw_threshold
            ] = 1.


    def detect_peaks(self, merge = None):

        peak_df = pd.DataFrame()

        merge = self.merge_peaks_by if merge is None else merge

        for channel, (channel_color, _) in self.channels.items():

            module_logger.debug(
                f"Running peak detection for {channel} channel"
            )
            peaks, properties = sig.find_peaks(
                self.data[channel_color],
                height = (self.peak_min_threshold, self.peak_max_threshold),
                distance = round(
                    self.peak_min_distance *
                    self.acquisition_rate
                ),
                prominence = (
                    self.peak_min_prominence,
                    self.peak_max_prominence,
                ),
                width = (
                    self.peak_min_width * self.acquisition_rate,
                    self.peak_max_width * self.acquisition_rate,
                ),
                rel_height = self.width_rel_height,
            )

            properties = pd.DataFrame.from_dict(properties)
            properties = properties.assign(barcode = channel == 'barcode')
            peak_df = peak_df.append(properties)

        # Converting ips values to int for indexing later on
        peak_df.assign(
            right_ips = round(peak_df.right_ips),
            left_ips = round(peak_df.left_ips),
        )
        peak_df = peak_df.astype({'left_ips': 'int32', 'right_ips': 'int32'})

        self.peak_df = peak_df

        if merge:

            self._merge_peaks(by = merge)


    def _merge_peaks(self, by = None):
        """
        Merges peaks if merge_peaks_distance is > 0.
        :param peak_df: DataFrame with peaks as called by detect_peaks()
        :return: List containing plug data.
        """

        peak_list = list()

        by = by or self.merge_peaks_by
        by_center = by == 'center'

        if not by_center or self.merge_peaks_distance:

            module_logger.info(
                'Merging %splugs %s' % (
                    'overlapping ' if by != 'center' else '',
                    'with centers closer than %.02f seconds' % (
                        self.merge_peaks_distance
                    )
                        if by == 'center' else
                    '',
                )
            )
            i_merge_dist = self.acquisition_rate * self.merge_peaks_distance

            merge_df = self.peak_df

            if by_center:

                merge_df = merge_df.assign(
                    peak_center = (
                        self.peak_df.left_ips +
                        (self.peak_df.right_ips - self.peak_df.left_ips) / 2
                    )
                )

            by_column = 'peak_center' if by_center else 'left_ips'
            merge_df = merge_df.sort_values(by = by_column)
            merge_df = merge_df.reset_index(drop = True)
            peaks = merge_df[by_column]
            rights = merge_df['right_ips']

            # Count through array
            i = 0
            while i < len(peaks):
                # Count neighborhood
                j = 0
                max_right = rights[i]
                while True:
                    if (
                        i + j >= len(peaks) or
                        (
                            peaks[i + j] - peaks[i] > i_merge_dist
                                if by_center else
                            peaks[i + j] > max_right
                        )
                    ):
                        # Merge from the left edge of the first plug (i)
                        # to the right base of the last plug
                        # to merge (i + j - 1)

                        left = min(merge_df.left_ips[i:i + j])
                        right = max(merge_df.right_ips[i:i + j])
                        peak_list.append(self.quantify_interval(left, right))

                        # Skip to the next unmerged plug
                        i += j
                        break
                    else:
                        max_right = max(max_right, rights[i + j])
                        j += 1

        else:

            module_logger.info(
                'Creating plug list without merging close plugs!'
            )

            for row in self.peak_df.sort_values(by = "left_ips"):

                peak_list.append(
                    self.quantify_interval(
                        row.left_ips,
                        row.right_ips
                    )
                )

         # Build plug_df DataFrame
        module_logger.debug('Building peak_df DataFrame')
        channels = [
            f"{str(key)}_peak_median"
            for key in self.config.channels.keys()
        ]

        self.peak_df = pd.DataFrame(
            peak_list,
            columns = ['start_time', 'end_time'] + channels,
        )


    def quantify_interval(self, start_index, end_index):
        """
        Calculates median for each acquired channel between two data points

        :param start_index: Index of the first datapoint in pmt_data.data
        :param end_index: Index of the last datapoint in pmt_data.data

        :return: List with start_index end_index and the channels according
        to the order of config.channels
        """

        i, j = start_index, end_index

        result = list()
        result.append(i / self.acquisition_rate + self.data.time.iloc[0])
        result.append(j / self.acquisition_rate + self.data.time.iloc[0])

        for _, (channel_color, _) in self.channels.items():

            result.append(self.data[channel_color][i:j].median())

        return result


    def _detection_issues_message(self):

        return (
            f"You may want to try:\n"
            f"\t- Increase `peak_max_width` (currently {self.peak_max_width})"
            f" if plugs are too short to be detected\n"
            f"\t- Decrease `width_rel_height` (currently "
            f"{self.width_rel_height}) if plugs are wider than tall\n"
            f"\t- Cut the data using the `cut` parameter (currently "
            f"{self.cut}) to remove parts from the beginning and end "
            f"which might look like a sample but in fact is not."
        )


#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# This file is part of the `plugy` python module
#
# Copyright
# 2018-2021
# EMBL & Heidelberg University
#
# Author(s): Dénes Türei (turei.denes@gmail.com)
#            Nicolas Peschke
#            Olga Ivanova
#
# Distributed under the GPLv3 License.
# See accompanying file LICENSE.txt or copy at
#     http://www.gnu.org/licenses/gpl-3.0.html
#
# Webpage: https://github.com/saezlab/plugy
#


import os
import sys
import re
import time
import pathlib as pl
import typing
import logging
import hashlib

from dataclasses import dataclass, field


module_logger = logging.getLogger(__name__)


@dataclass
class PlugyConfig(object):

    # File Paths
    input_dir: str = '.'
    pmt_file: typing.Union[pl.Path, str, re.Pattern] = \
        re.compile(
            r'(?:fluor|exp)(?:[-\w ]*)?'
            r'(?:\.txt)(?:\.gz)?',
            re.IGNORECASE
        )
    seq_file: typing.Union[pl.Path, str, re.Pattern] = \
        re.compile(r'seq(?:[-\w ]*)?(?:\.[tc]sv)?', re.IGNORECASE)
    channel_file: typing.Union[pl.Path, str, re.Pattern] = \
        re.compile(r'channel(?:[-\w ]*)?(?:\.[tc]sv)?', re.IGNORECASE)
    result_base_dir: typing.Union[pl.Path, str] = None
    result_subdirs: bool = False
    timestamp_result_subdirs: bool = False

    ## General config
    name: str = None
    # file type for saving the figures
    figure_export_file_type: str = 'png'
    # colors for representing channels on the figures
    colors: dict = field(
        default_factory = lambda: {
            'green': '#5D9731',
            'uv': '#3A73BA',
            'orange': '#F68026',
        }
    )
    # run the entire workflow
    run: bool = False
    # run the initial steps: set up the objects and load the data
    init: bool = None
    # run the plug detection
    plugs: bool = None
    # run the sample detection
    samples: bool = None
    # run the quality control
    qc: bool = None
    # run the analysis
    analysis: bool = None
    # label of the control channels in the channel layout
    control_label: typing.Union[set, str] = 'FS'
    medium_control_label: typing.Union[set, str] = 'FS'
    negative_control_label: typing.Union[set, str] = field(
        default_factory = lambda: {'neg-ctrl', 'neg_ctrl'}
    )
    positive_control_label: typing.Union[set, str] = field(
        default_factory = lambda: {'pos-ctrl', 'pos_ctrl'}
    )
    # do not raise an error if the quality control fails
    # but proceed with quantification and visualization
    # of the results
    ignore_qc_result: bool = True
    # whether to write log messages also to the STDOUT
    # or only to the log file
    log_to_stdout: bool = True

    # sequence settings
    allow_lt4_valves: bool = False

    ## PMT configuration
    # function, name and column index of the channels
    channels: dict = field(
        default_factory = lambda: {
            'barcode': ('uv', 3),
            'control': ('orange', 2),
            'readout': ('green', 1),
        }
    )
    # PMT acquisition rate in Hz
    acquisition_rate: int = 300
    # crop the data by removing the segments before the first
    # and after the second element of this tuple (in seconds)
    cut: tuple = (None, None)

    # set the acquisition times by a linear interpolation
    # based on the `acquisition_rate`
    correct_acquisition_time: bool = True
    # set the value of the channels listed here to zero
    ignore_channels: set = field(default_factory = set)
    # scale certain channels by a constant factor
    fake_gains: dict = field(default_factory = dict)
    # a default scaling factor for the remaining channels
    fake_gain_default: float = 1.0
    # I don't know what this intended to be,
    # raises NotImplementedError
    fake_gain_adaptive: bool = False
    # above this value set the value of the barcode channel to 1.0
    barcode_raw_threshold: float = None

    # parameters for the peak detection algorithm;
    # we use a generic signal processing algorithm
    # from scipy.signal.find_peaks
    auto_detect_cycles: bool = True
    peak_min_threshold: float = 0.05
    peak_max_threshold: float = 2.0
    peak_min_distance: float = 0.03
    peak_min_prominence: float = 0
    peak_max_prominence: float = 10
    peak_min_width: float = 0.5
    peak_max_width: float = 2.5
    width_rel_height: float = 0.5
    # max distance between centers of the peaks to merge
    # if we merge the peaks by their centers
    merge_peaks_distance: float = 0.2
    # alternative is to merge them by 'center', but unless for
    # some weird reason supposedly distinct peaks overlap
    # merging by overlap is a superior method
    merge_peaks_by: str = 'overlap'
    n_bc_adjacent_discards: int = 1
    # lowest number of barcode plugs separating two cycles
    # barcode segments with fewer plugs separate samples within cycles
    min_end_cycle_barcodes: int = 12
    # lowest number of barcode plugs separating 2 samples
    # if fewer barcode plugs between sample plugs, the sample
    # counter won't be increased
    min_between_samples_barcodes: int = 2
    # lowest number of plugs in a sample
    # if barcode comes without reaching yet this threshold,
    # the sample counter won't be increased
    min_plugs_in_sample: int = 3

    # if False, the workflow stops after plug identification and
    # quantification, sample and barcode plugs won't be distinguished
    has_barcode: bool = True
    # if False, plug detection stops after identification of barcode plugs
    # and no sample and cycle numbers will be assigned to the plugs
    has_samples_cycles: bool = True
    # the number of samples within one cycle; no need to provide if sequence
    # file is available; otherwise it helps in evaluating barcode detection
    # methods even if the sequence is unknoen
    samples_per_cycle: int = None
    # The experiment has positive (and optionally negative) control samples.
    # If False, the tasks which require those samples will be skipped.
    has_controls: bool = True
    # the name of a method for identifying barcode plugs:
    # * `simple`: the simplest method, works with not too noisy data:
    #    plugs are barcode if the channel of the blue value is the highest
    # * `adaptive`: change the value of the `blue_highest_times`
    #    parameter to find the one giving the best result
    barcoding_method: str = 'simple'
    # override parameters for the selected barcode method
    barcoding_param: dict = field(default_factory = dict)
    # default parameters for the barcode methods
    barcoding_param_defaults: dict = field(
        default_factory = lambda: {
            'simple': {
                # a scaling factor for the `blue_highest` method: the blue
                # channel must be at least this times higher than the control
                # channel for barcode plugs
                'times': 1.0,
            },
            'adaptive': {
                'adaptive_method': 'simple',
                'higher_threshold_factor': 1.,
                'thresholding_method': 'local',
                'block_size': 7,
            }
        }
    )

    append: list = field(default_factory = list)

    # Analysis
    normalize_using_control: bool = True
    normalize_using_media_control_lin_reg: bool = True
    readout_column: str = "readout_peak_median"
    readout_analysis_column: str = "readout_media_norm_z_score"

    # Plotting config
    seaborn_context: str = "notebook"
    seaborn_context_dict: dict = field(
        default_factory = lambda: {
            'grid.linewidth': .9,
        }
    )
    seaborn_style: str = "whitegrid"
    seaborn_style_dict: dict = field(
        default_factory = lambda: {
            'axes.edgecolor': '#000000',
            'grid.color': '#000000',
        }
    )
    font_scale: typing.Union[float, int] = 2
    scatter_dot_size: typing.Union[float, int] = 10
    plot_git_caption: bool = True

    heatmap_second_scale: str = 'pos_ctrl'
    heatmap_override_scale: tuple = None
    heatmap_override_second_scale: tuple = None

    # shorten drug names on certain plots
    short_labels: bool = True

    # enable/disable figure titles
    figure_titles: bool = True

    # QC
    contamination_threshold: float = 0.03

    # Statistics
    alpha: float = 0.05

    # Palette for plotting
    palette: tuple = (
        '#7264B9',
        '#5B205F',
        '#9E1639',
        '#ED0772',
        '#D22027',
    )

    def __post_init__(self):
        # Creating result dir for each individual run


        self._set_result_dir()
        self.start_logging()
        self._set_paths()
        self._set_name()


    def _set_result_dir(self):

        self.result_base_dir = (
            pl.Path(self.result_base_dir)
                if isinstance(self.result_base_dir, str) else
            self.result_base_dir
        )

        self.result_dir = (
            self.result_base_dir or
            pl.Path.cwd().joinpath('results')
        )

        if self.result_subdirs:

            current_time = time.strftime("%Y%m%d_%H%M")

            subdir = (
                f"{os.path.splitext(self.pmt_file.name)[0]}_{current_time}"
                    if self.timestamp_result_subdirs else
                self.name
            )

            self.result_dir = self.result_dir.joinpath(
                self.result_dir,
                subdir,
            )

        os.makedirs(str(self.result_dir), exist_ok = True)


    def _set_paths(self):

        for attr in (
            'pmt_file',
            'channel_file',
            'seq_file',
        ):

            path = getattr(self, attr)

            if not isinstance(path, pl.Path):

                path = self._find_file(path = path, in_dir = self.input_dir)

                if not os.path.exists(path):

                    msg = 'File not found: %s; `%s`' % (attr, path)
                    module_logger.critical(msg)
                    module_logger.critical(
                        'Files in directory `%s`: %s' % (
                            self.input_dir,
                            ', '.join(os.listdir(self.input_dir))
                        )
                    )
                    path = None

                    if attr == 'pmt_file':

                        raise FileNotFoundError(msg)

                    else:

                        setattr(self, attr, None)

                else:

                    msg = 'Input file `%s`: %s' % (attr, path)
                    module_logger.info(msg)
                    path = pl.Path(path)

                setattr(self, attr, path)


    @staticmethod
    def _find_file(path, in_dir):

        if hasattr(path, 'pattern'):

            subdir, fname = (
                ('', path.pattern)
                    if sys.platform.lower().startswith('win') else
                os.path.split(path.pattern)
            )

            in_dir = in_dir + subdir

            for f in os.listdir(in_dir):

                this_path = os.path.join(subdir, f)

                if path.match(this_path):

                    path = os.path.join(in_dir, this_path)
                    break

        return path.pattern if hasattr(path, 'pattern') else path


    def _set_name(self):

        self.identity = hashlib.md5(
            ';'.join(
                (
                    str(self.pmt_file),
                    str(self.seq_file),
                    str(self.channel_file),
                )
            ).encode('utf-8')
        ).hexdigest()[:7]

        self.name = (
            self.name or
            '%s-%s' % (
                os.path.splitext(self.pmt_file.name)[0],
                self.identity,
            )
        )


    def start_logging(self):
        """
        Starts logging to STDOUT and to a log file in the result directory
        :param config: PlugyConfig object to retrieve the result_dir from
        :return: None
        """

        logging.shutdown()

        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
            datefmt = "%d.%m.%y %H:%M:%S",
        )

        logger = logging.getLogger()
        logger.handlers = []
        logger.setLevel(logging.DEBUG)

        if self.log_to_stdout:

            stream_handler = logging.StreamHandler(stream=sys.stdout)
            stream_handler.setLevel(logging.INFO)
            stream_handler.setFormatter(formatter)
            logger.addHandler(stream_handler)

        file_handler = logging.FileHandler(
            self.result_dir.joinpath('plugy.log'),
            mode = 'a'
        )
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)

        logger.addHandler(file_handler)


    @property
    def channel_names(self):

        return dict(
            (key, value[0])
            for key, value in self.channels.items()
        )


    def channel_color(self, channel):

        return (
            self.colors[self.channels[channel][0]]
                if channel in self.channels else
            '#000000'
        )
